/*
 * process.c
 *
 * Copyright (C) 2006  Insigme Co., Ltd
 *
 * This software has been developed while working on the Linux Unified Kernel
 * project (http://www.longene.org) in the Insigma Research Institute,
 * which is a subdivision of Insigma Co., Ltd (http://www.insigma.com.cn).
 *
 * The project is sponsored by Insigma Co., Ltd.
 *
 * The authors can be reached at linux@insigma.com.cn.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of  the GNU General  Public License as published by the
 * Free Software Foundation; either version 2 of the  License, or (at your
 * option) any later version.
 *
 * Revision History:
 *   Jan 2006 - Created.
 */

/*
 * process.c:
 * Refered to ReactOS code
 */
#include "win32_process.h"
#include <linux/mm.h>
#include <linux/syscalls.h>
#include <asm/uaccess.h>
#include <linux/sched.h>
#include <linux/fs_struct.h>
#include <linux/dcache.h>
#include "ke.h"
#include "win32.h"
#include "unistr.h"
#include "thread.h"
#include "event.h"
#include "object.h"
#include "process.h"
#include "section.h"
#include "handle.h"
#include "objwait.h"
#include "wineserver/info.h"

#ifdef CONFIG_UNIFIED_KERNEL

POBJECT_TYPE process_object_type = NULL;
EXPORT_SYMBOL(process_object_type);

extern char ntdll_name[];

extern long do_fork_from_task(struct task_struct *ptsk,
		unsigned long process_flags,
		unsigned long clone_flags,
		unsigned long stack_start,
		struct pt_regs *regs,
		unsigned long stack_size,
		int __user *parent_tidptr,
		int __user *child_tidptr);
extern void do_exit_task(struct task_struct *tsk, long code);
extern void do_group_exit(int exit_code);
extern NTSTATUS STDCALL map_system_dll(struct task_struct *tsk, char *name,
		unsigned long *ntdll_load_addr, unsigned long *interp_load_addr);

extern NTSTATUS STDCALL create_process_space(struct eprocess *, struct win32_section *);

extern void *mem_alloc(size_t size);

#ifdef SERVER_PM
/* process startup info */
struct startup_info
{
    struct object       obj;          /* object header */
    obj_handle_t        hstdin;       /* handle for stdin */
    obj_handle_t        hstdout;      /* handle for stdout */
    obj_handle_t        hstderr;      /* handle for stderr */
    struct file        *exe_file;     /* file handle for main exe */
    struct process     *process;      /* created process */
    data_size_t         data_size;    /* size of startup data */
    void               *data;         /* data for startup info */
};
#endif

/* create a new w32process struct, but no w32thread will be created in this function */
struct w32process *create_w32process(struct eprocess *eproc) /* TBD, D.M. */
{
	struct w32process *w32proc = NULL;

	if ( !(w32proc = kmalloc(sizeof(struct w32process), GFP_KERNEL)) )
		return NULL;
	memset(w32proc, 0, sizeof(struct w32process));

	w32proc->parent          = NULL;
	w32proc->debugger        = NULL;
	w32proc->sigkill_timeout = NULL;
	w32proc->running_threads = 0;
	w32proc->suspend         = 0;
	w32proc->is_system       = 0;
	w32proc->create_flags    = 0;
	w32proc->console         = NULL;
	w32proc->startup_state   = STARTUP_IN_PROGRESS;
	w32proc->startup_info    = NULL;
	w32proc->idle_event      = NULL;
	w32proc->queue           = NULL;
	w32proc->ldt_copy        = NULL;
	w32proc->winstation      = 0;
	w32proc->desktop         = 0;
	w32proc->trace_data      = 0;

	INIT_LIST_HEAD(&w32proc->locks);
	INIT_LIST_HEAD(&w32proc->classes);
	INIT_LIST_HEAD(&w32proc->dlls);
	w32proc->eprocess = eproc;

	w32proc->start_time = current_time;
	w32proc->end_time = 0;

#if 0
#ifndef UNIFIED_KERNEL
	if (!(process->id = process->group_id = alloc_ptid( process )))
	{
		close( fd );
		goto error;
	}
#else
	process->id = process->group_id = 0;
#endif
	if (!(process->msg_fd = create_anonymous_fd( &process_fd_ops, fd, &process->obj, 0 ))) goto error;

	/* create the handle table */
	if (!parent_thread)
	{
		process->handles = alloc_handle_table( process, 0 );
		process->token = token_create_admin();
	}
	else
	{
		struct process *parent = parent_thread->process;
		process->parent = (struct process *)grab_object( parent );
		process->handles = inherit_all ? copy_handle_table( process, parent )
			: alloc_handle_table( process, 0 );
		/* Note: for security reasons, starting a new process does not attempt
		 * to use the current impersonation token for the new process */
		process->token = token_duplicate( parent->token, TRUE, 0 );
	}
	if (!process->handles || !process->token) goto error;

	/* create the main thread */
	if (pipe( request_pipe ) == -1)
	{
		file_set_error();
		goto error;
	}
	fprintf(stderr, "%s \n", __func__);
	if (send_client_fd( process, request_pipe[1], 0 ) == -1)
	{
		close( request_pipe[0] );
		close( request_pipe[1] );
		goto error;
	}
	close( request_pipe[1] );

	if (!(thread = create_thread( request_pipe[0], process ))) goto error;

	set_fd_events( process->msg_fd, POLLIN );  /* start listening to events */
	release_object( process );
#endif

	return w32proc;
#if 0
error:
	kfree( w32proc );
	return NULL;
#endif
}

/*
 * poll the state of a process for WaitFor*() functions
 * - if signalled, sets to non-signalled before returning
 */
int poll_process(struct wait_table_entry *wte)
{
	struct eprocess * process = (struct eprocess *)(wte->wte_obj);

	return (process->pcb.state == PROCESS_STATE_ACTIVE) ? POLL_NOTSIG : POLL_SIG;
} /* end poll_process() */

VOID delete_process(PVOID Object)
{
	struct eprocess *process = Object;

	ktrace("delete_process\n");

	if (process->unique_processid)
		delete_cid_handle(process->unique_processid, process_object_type);

	if (process->ep_nls)
		unload_nls(process->ep_nls);
}

/* initialize kprocess */
void kprocess_init(struct kprocess *process, char prio,
		unsigned long affinity, physical_address_t dir_table_base)
{
	ktrace("kprocess_init\n");

	INIT_DISP_HEADER(&process->header, process_object, sizeof(struct kprocess), false);

	/* Initialize Scheduler Data, Disable Alignment Faults and Set the PDE */
	process->affinity = affinity;
	process->base_priority = prio;
	process->quantum_reset = 6;
	process->directory_table_base = dir_table_base;
	process->auto_alignment = true;
	process->state = PROCESS_STATE_ACTIVE;

	/* Initialize the Thread List */
	INIT_LIST_HEAD(&process->thread_list_head);
} /* end kprocess_init */

/* initialize eprocess */
void eprocess_init(struct eprocess *process)
{
	physical_address_t      dir_table_base = { .quad = 0LL };

	ktrace("eprocess_init\n");

	process->ep_nls = load_nls("cp936");

	INIT_LIST_HEAD(&process->thread_list_head);
	rwlock_init(&process->ep_lock);

	process->debug_port = NULL;             /* FIXME */
	process->exception_port = NULL;         /* FIXME */
	process->section_base_address = 0;      /* FIXME */
	/* FIXME: event_init(&Process->LockEvent, SynchronizationEvent, FALSE); */
	/* FIXME: process->object_table = process->default_object_table; */
	/* FIXME: Status = PspInitializeProcessSecurity(Process, pParentProcess); */

	/* init handle table */
	create_handle_table(NULL, FALSE, process);

	/* 
	 * FIXME:
	 * affinity == active processors, used 1 here
	 * directory_table_base, used 0 here 
	 */
	kprocess_init(&process->pcb, PROCESS_PRIO_NORMAL, 1, dir_table_base);

	INIT_LIST_HEAD(&process->ep_reserved_head);
	INIT_LIST_HEAD(&process->ep_mapped_head);
	
	process->watch_fd = -1;
	process->watch_thread = 0;
	process->win32process = create_w32process(process);
	process->ep_handle_info_table = alloc_handle_info_table();
}


/*
 * create_process
 */
NTSTATUS STDCALL create_process(OUT PHANDLE ProcessHandle,
                 IN ACCESS_MASK DesiredAccess,
                 IN POBJECT_ATTRIBUTES ObjectAttributes  OPTIONAL,
                 IN HANDLE ParentProcessHandle  OPTIONAL,
                 IN BOOLEAN InheritObjectTable,
                 IN HANDLE SectionHandle  OPTIONAL,
                 IN HANDLE DebugPort  OPTIONAL,
                 IN HANDLE ExceptionPort  OPTIONAL)
{
	HANDLE hProcess = NULL;
	struct eprocess *child_eprocess;
	struct eprocess *parent_eprocess = NULL;
	NTSTATUS Status = STATUS_SUCCESS;
	char *system_dll_name = ntdll_name;
	long	cpid;
	struct task_struct	*child_task = NULL, *parent_task = NULL;
	struct ethread *current_ethread = get_current_ethread();
	struct ethread *parent_ethread = NULL;
	struct ethread *child_ethread = NULL;
	struct ktrap_frame	*trap_frame = NULL;
	struct win32_section	*section = NULL;
	unsigned long system_dll_load_addr, interp_load_addr;

	if (!ProcessHandle)
		return STATUS_INVALID_PARAMETER;

	if (!current_ethread)
		return STATUS_INVALID_PARAMETER;

    	/* Reference the Parent if there is one */
	if (ParentProcessHandle == NtCurrentProcess()) {
		parent_eprocess = current_ethread->threads_process;
		parent_ethread = current_ethread;
		ref_object((PVOID)parent_eprocess);
	}
	else {
		Status = ref_object_by_handle(ParentProcessHandle,
				PROCESS_ALL_ACCESS,
				process_object_type,
				KernelMode,
				(PVOID *)&parent_eprocess,
				NULL);
		if (!NT_SUCCESS(Status)) {
			return Status;
		}
		parent_ethread = get_first_thread(parent_eprocess);
	}

	trap_frame = parent_ethread->tcb.trap_frame;
	parent_task = parent_ethread->et_task;

	cpid = do_fork_from_task(parent_task, CREATE_PROCESS,
			SIGCHLD | CLONE_VM | CLONE_FS | CLONE_FILES | CLONE_SIGHAND,	/* SIGCHLD ????? */
			trap_frame->esp, (struct pt_regs *)trap_frame, 0, NULL, NULL);

	if (parent_eprocess)
		deref_object((PVOID)parent_eprocess);

	if (cpid < 0) {
		Status = STATUS_INVALID_PARAMETER;
		goto out;
	}

	child_task = find_task_by_vpid(cpid);

	/* Create EPROCESS */
	Status = create_object(KernelMode,
			process_object_type,
			ObjectAttributes,
			KernelMode,
			NULL,
			sizeof(struct eprocess),
			0,
			0,
			(PVOID *)&child_eprocess);
	if (!NT_SUCCESS(Status))
		goto cleanup_child_task;

	eprocess_init(child_eprocess);

	/* Insert Process into Handle Table */
	Status = insert_object(child_eprocess,
			NULL,
			DesiredAccess,
			0,
			NULL,
			&hProcess);
	if (!NT_SUCCESS(Status))
		goto cleanup_child_eprocess;

	/* allocate a Win32 thread object */
	Status = create_object(KernelMode,
			thread_object_type,
			ObjectAttributes,
			KernelMode,
			NULL,
			sizeof(struct ethread),
			0,
			0,
			(PVOID *)&child_ethread);
	if (!NT_SUCCESS(Status))
		goto cleanup_proc_handle;

	ethread_init(child_ethread, child_eprocess, child_task);

	child_eprocess->fork_in_progress = child_ethread;

	/* Inherit stuff from the Parent since we now have the object created */
	if (parent_eprocess) {
		child_eprocess->inherited_from_unique_pid = parent_eprocess->unique_processid;
		child_eprocess->session = parent_eprocess->session;
	}

	/* FIXME: Set up the Quota Block from the Parent
	 *	 PspInheritQuota(Parent, child_eprocess); 
	 */

	/* FIXME: Set up Dos Device Map from the Parent
	 *  ObInheritDeviceMap(Parent, child_eprocess) 
	 */

	/* Set the Process' LPC Ports */

	/* Setup the Lock Event */
	event_init(&child_eprocess->lock_event, synchronization_event, FALSE);

	/* Add the Section */
	if (SectionHandle) {
		Status = ref_object_by_handle(SectionHandle,
				0,
				section_object_type,
				KernelMode,
				(PVOID *)&section,
				NULL);
		if (!NT_SUCCESS(Status))
			goto cleanup_child_ethread;
	}

	/* Create the Process' Address Space */
	Status = create_process_space(child_eprocess, section);
	if (!NT_SUCCESS(Status))
		goto cleanup_section;

	/* Do what exec should do */
	flush_old_exec_from_task(child_task);

	if (section) {
		/* Map the System Dll */
		map_system_dll(child_task, system_dll_name,
				&system_dll_load_addr, &interp_load_addr);
	}

	child_eprocess->unique_processid = create_cid_handle(child_eprocess, process_object_type);
	if (!(child_eprocess->unique_processid))
		goto cleanup_section;

	/* Create PEB only for User-Mode Processes */
	if (parent_eprocess) {
		Status = create_peb(child_eprocess);
		if (!NT_SUCCESS(Status))
			goto cleanup_section;
	}

	/* Let's take advantage of this time to kill the reference too */
	parent_eprocess = NULL;

    	/* Set the Creation Time */
	if (copy_to_user(ProcessHandle, &hProcess, sizeof(hProcess))) {
		Status = STATUS_UNSUCCESSFUL;
		goto cleanup_section;
	}

	deref_object(child_eprocess);
	deref_object(child_ethread);

    	return STATUS_SUCCESS;

cleanup_section:
	deref_object((PVOID)section);

cleanup_child_ethread:
	deref_object((PVOID)child_ethread);

cleanup_proc_handle:
	if (hProcess)
		NtClose(hProcess);

cleanup_child_eprocess:
	deref_object(child_eprocess);

cleanup_child_task:
	do_exit_task(child_task, 0);

out:
    	return Status;
} /* end create_process */

NTSTATUS SERVICECALL
NtCreateProcess(OUT PHANDLE ProcessHandle,
                IN ACCESS_MASK DesiredAccess,
                IN POBJECT_ATTRIBUTES ObjectAttributes  OPTIONAL,
                IN HANDLE ParentProcessHandle,
                IN BOOLEAN InheritObjectTable,
                IN HANDLE SectionHandle  OPTIONAL,
                IN HANDLE DebugPort  OPTIONAL,
                IN HANDLE ExceptionPort  OPTIONAL)
{
	return ParentProcessHandle
		? create_process(ProcessHandle,
				DesiredAccess,
				ObjectAttributes,
				ParentProcessHandle,
				InheritObjectTable,
				SectionHandle,
				DebugPort,
				ExceptionPort)
		: STATUS_INVALID_PARAMETER;
} /* end NtCreateProcess */
EXPORT_SYMBOL(NtCreateProcess);

NTSTATUS SERVICECALL
NtOpenProcess(OUT PHANDLE ProcessHandle, 
	      IN ACCESS_MASK DesiredAccess, 
	      IN POBJECT_ATTRIBUTES ObjectAttributes, 
	      IN PCLIENT_ID ClientId)
{
	/* FIXME: combine Ke and Ex into the one */
	KPROCESSOR_MODE pre_mode = (KPROCESSOR_MODE)get_pre_mode();
	struct eprocess *process = NULL;
	struct ethread *thread = NULL;
	NTSTATUS status = STATUS_INVALID_PARAMETER;
	CLIENT_ID cid;
	
	ktrace("NtOpenProcess\n");

	if (ObjectAttributes->ObjectName) {
		return open_object_by_name(ObjectAttributes, 
				process_object_type, 
				NULL, 
				pre_mode, 
				DesiredAccess, 
				NULL, 
				ProcessHandle);
	} else if (ClientId) {
		if (copy_from_user(&cid, ClientId, sizeof(cid)))
			return STATUS_INVALID_ADDRESS;
		if (cid.UniqueProcess) {
			if (lookup_process_by_pid(cid.UniqueProcess, &process))
				return status;
		} else if (cid.UniqueThread) {
			if (lookup_process_thread_by_cid(&cid, &process, &thread))
				return status;
		} else
			return status;
		
		status = open_object_by_pointer(process, 
				ObjectAttributes->Attributes, 
				NULL, 
				DesiredAccess, 
				process_object_type, 
				pre_mode, 
				ProcessHandle);

		if (thread)
			deref_object(thread);
		
		deref_object(process);
	}
	return status;
}
EXPORT_SYMBOL(NtOpenProcess);

NTSTATUS SERVICECALL
NtTerminateProcess(IN HANDLE ProcessHandle,
		IN NTSTATUS ExitStatus)
{
	struct eprocess *process;
	struct ethread *cur_thread;
	NTSTATUS status;

	ktrace("NtTerminateProcess\n");

	if ((status = ref_object_by_handle(ProcessHandle ? ProcessHandle : NtCurrentProcess(),
				PROCESS_TERMINATE, 
				process_object_type, 
				get_pre_mode(), 
				(PVOID *) &process, 
				NULL)))
		return status;

	cur_thread = (struct ethread *) get_current_ethread();

	lock_process(process);

	if (process->exit_time.quad) {
		unlock_process(process);
		deref_object(process);
		return STATUS_PROCESS_IS_TERMINATING;
	}

	query_sys_time(&process->exit_time);
	process->exit_status = (unsigned long)ExitStatus;

	unlock_process(process);
	deref_object(process);

	if (process == get_current_eprocess()) {
		cur_thread->exit_status = ExitStatus;
		do_group_exit((ExitStatus & 0xff) << 8);
	} else {
		struct ethread *first_thread = get_first_thread(process);

		first_thread->exit_status = ExitStatus;
		send_sig_info(SIGKILL, SEND_SIG_FORCED, first_thread->et_task->group_leader);
	}

	return STATUS_SUCCESS;
}
EXPORT_SYMBOL(NtTerminateProcess);

/* copy unicode string to PPB */
static inline int
copy_param_str(void *ptr, PUNICODE_STRING dst, PUNICODE_STRING src)
{
	int		ret = 0;

	dst->Length = src->Length;
	dst->MaximumLength = src->MaximumLength;
	dst->Buffer = (PWSTR)ptr;
	if (src->Length)
		ret = copy_to_user(dst->Buffer, src->Buffer, src->Length + sizeof(wchar_t));

	return ret ? -EFAULT : ALIGN_TO_LONG(dst->MaximumLength);
}

static int copy_env(struct nls_table *nls, PWSTR *pwcs, char *src, int len)
{
	char		*ksrc, *p;
	PWSTR	wcs = *pwcs;
	wchar_t	*wc;
	int		ret = -ENOMEM;
	int		i = 0;

	if (wcs)
		return -EINVAL;

	ksrc = kmalloc(len + 1, GFP_KERNEL);
	if (!ksrc)
		return ret;

	ret = -EFAULT;
	if (copy_from_user(ksrc, src, len))
		goto out_free_src;
	ksrc[len++] = 0;

	wcs = kmalloc(len * sizeof(wchar_t), GFP_KERNEL);
	ret = -ENOMEM;
	if (!wcs)
		goto out_free_src;

	ret = -EINVAL;
	p = ksrc;
	wc = wcs;
	while (p < ksrc + len) {
		i = nls->char2uni(p, sizeof(wchar_t), wc++);
		if (i < 0)
			goto out_free_wcs;
		p += i;
	}

	ret = (char *)wc - (char *)wcs;
	*pwcs = wcs;
	goto out_free_src;

out_free_wcs:
	kfree(wcs);
out_free_src:
	kfree(ksrc);

	return ret;
}

long wcslen_user(const wchar_t *s)
{
	long res = 0;
	wchar_t c = 0;

	for (;;) {
		if (get_user(c, s))
			return 0;
		if (!c)
			return res + sizeof(wchar_t);
		
		res += sizeof(wchar_t);
		s++;
	}
}
EXPORT_SYMBOL(wcslen_user);

static int copy_cmdline(struct nls_table *nls, PUNICODE_STRING dst, int argc, char **argv, int total_len)
{
	int	i, n, len;
	char	*p, *kargv;
	int	ret = -ENOMEM;
	wchar_t	*wc;

	if (!(kargv = kmalloc(total_len, GFP_KERNEL)))
		return ret;

	dst->Length = 0;
	dst->MaximumLength = total_len * sizeof(wchar_t) + argc * 2 * sizeof(wchar_t);
	dst->Buffer = kmalloc(dst->MaximumLength, GFP_KERNEL);
	if (!dst->Buffer)
		goto out_free_src;

	wc = dst->Buffer;

	for (i = 0; i < argc; i++) {
		len = (i == argc - 1) ? strlen_user(argv[i]) : argv[i + 1] - argv[i];
		ret = -EFAULT;
		if (copy_from_user(kargv, argv[i], len))
			goto out_free_src;
		p = kargv;

		if (*p != '"') {
			dst->Length += sizeof(wchar_t);
			*wc++ = L'"';
		}
		while (p < kargv + len - 1) {
			n = nls->char2uni(p, sizeof(wchar_t), wc++);
			if (n < 0) {
				ret = -EINVAL;
				goto out_free_dst;
			}
			p += n;
			dst->Length += sizeof(wchar_t);
		}

		if (*kargv != '"') {
			dst->Length += sizeof(wchar_t);
			*wc++ = L'"';
		}
		if (i == argc - 1) {
			*wc++ = L'\0';
			dst->Length += sizeof(wchar_t);
		} else {
			*wc++ = L' ';
			dst->Length += sizeof(wchar_t);
		}
	}

	ret = 0;
	goto out_free_src;

out_free_dst:
	kfree(dst->Buffer);
	memset(dst, 0, sizeof(UNICODE_STRING));
out_free_src:
	kfree(kargv);
	return ret;
}

/* create ppb */
NTSTATUS
create_ppb(PRTL_USER_PROCESS_PARAMETERS *ppb_res,
		struct eprocess *process,
		struct linux_binprm *bprm,
		char *image_name,
		char *dll_path,
		char *current_dir,
		PWSTR environ,
		char *window_title,
		char *desktop_info,
		char *shell_info,
		char *rt_info)
{
	PRTL_USER_PROCESS_PARAMETERS ppb = NULL;
	HANDLE current_dir_handle;
	HANDLE console_handle;
	int		i, env_size, len, brk_size;
	int		ret = -EINVAL;
	unsigned long	console_flags;
	unsigned long	addr;
#ifndef EXE_SO
	unsigned long	pos = bprm->p;
#else
	unsigned long	pos = current->mm->arg_start;
#endif
	unsigned long	argv_start, envp_start;
	unsigned long	brk_res;
	char	**argv, **envp;
	char	*p, *cdir;
	char	*native_app = NULL, *native_cmdline = NULL;
	char	*desktop = NULL;
	int	cdir_len;
	void	*ptr;
	UNICODE_STRING	dll_path_uni = {0, 2, NULL};
	UNICODE_STRING	current_dir_uni;
	UNICODE_STRING	window_title_uni = {0, 2, NULL};
	UNICODE_STRING	desktop_info_uni = {0, 2, NULL};
	UNICODE_STRING	shell_info_uni = {0, 2, NULL};
	UNICODE_STRING	rt_info_uni = {0, 2, NULL};
	UNICODE_STRING	image_name_uni;
	UNICODE_STRING	cmd_line_uni;
	PWSTR	kenv = NULL;
	struct nls_table	*nls = process->ep_nls;

	cdir = (char *)kmalloc(0x400, GFP_KERNEL);
	argv = (char **)kmalloc(bprm->argc * sizeof(char *), GFP_KERNEL);
	envp = (char **)kmalloc(bprm->envc * sizeof(char *), GFP_KERNEL);

	argv_start = pos;
	for (i = 0; i < bprm->argc; i++) {
		argv[i] = (char *)pos;
		pos += strlen_user(argv[i]);
	}

	envp_start = pos;
	for (i = 0; i < bprm->envc; i++) {
		envp[i] = (char *)pos;
		pos += strlen_user(envp[i]);
	}

	for (i = 0; i < bprm->envc; i++) {
		if (!native_app && !strncmp(envp[i], "NATIVEAPP=", strlen("NATIVEAPP="))) {
			len = strlen_user(envp[i]);
			native_app = (char *)kmalloc(len, GFP_KERNEL);
			if (!native_app) {
				ret = -ENOMEM;
				if (native_cmdline)
					kfree(native_cmdline);
				goto out_free_envp;
			}
			if (copy_from_user(native_app, envp[i] + strlen("NATIVEAPP="), 
				len - strlen("NATIVEAPP="))) {
				ret = -EFAULT;
				kfree(native_app);
				if (native_cmdline)
					kfree(native_cmdline);
				goto out_free_envp;
			}
			continue;
		}

		if (!native_cmdline && !strncmp(envp[i], "NATIVECMDLINE=", 
			strlen("NATIVECMDLINE="))) {
			len = strlen_user(envp[i]);
			native_cmdline = (char *)kmalloc(len, GFP_KERNEL);
			if (!native_cmdline) {
				ret = -ENOMEM;
				if (native_app)
					kfree(native_app);
				goto out_free_envp;
			}
			if (copy_from_user(native_cmdline, envp[i] + strlen("NATIVECMDLINE="), 
				len - strlen("NATIVECMDLINE="))) {
				ret = -EFAULT;
				kfree(native_cmdline);
				if (native_app)
					kfree(native_app);
				goto out_free_envp;
			}
			continue;
		}

		if (!desktop && !strncmp(envp[i], "DESKTOP=", strlen("DESKTOP="))) {
			len = strlen_user(envp[i]);
			desktop = (char *)kmalloc(len, GFP_KERNEL);
			if (!desktop) {
				ret = -ENOMEM;
				goto out_free_envp;
			}
			if (copy_from_user(desktop, envp[i] + strlen("DESKTOP="), 
				len - strlen("DESKTOP="))) {
				ret = -EFAULT;
				kfree(desktop);
				goto out_free_envp;
			}
			continue;
		}
	}

	if ((ret = copy_cmdline(nls, &cmd_line_uni, bprm->argc, argv, pos - argv_start)))
		goto out_free_envp;

	if (native_cmdline) {
		str2unistr(nls, &cmd_line_uni, native_cmdline);
		kfree(native_cmdline);
	}

	/* copy environment from user to kernel */
	env_size = copy_env(nls, &kenv, (char *)envp_start, pos - envp_start);
	if ((ret = env_size) < 0)
		goto out_free_cmd;

	if (*image_name == '/')
		strcpy(cdir, image_name);
	else {
		for (i = 0; i < bprm->envc; i++) {
			if (strncmp(envp[i], "PWD", strlen("PWD")))
				continue;

			p = envp[i] + strlen("PWD");
			while (*p && *p != '=')
				p++;

			if (*p)
				current_dir = p + 1;

			cdir_len = strlen_user(current_dir) - 1;
			ret = -EFAULT;
			if (copy_from_user(cdir, current_dir, cdir_len))
				goto out_free_kenv;
			if (cdir[cdir_len - 1] != '/')
				cdir[cdir_len++] = '/';

			if (*image_name == '.' && image_name[1] == '/')
				strcpy(cdir + cdir_len, image_name + 2);
			else
				strcpy(cdir + cdir_len, image_name);
			strcpy(image_name, cdir);
		}
	}

	p = strrchr(cdir, '/');
	if (!p)
		goto out_free_kenv;
	*++p = '\0';

	len = p - cdir;
	current_dir = (char *)((bprm->p - len - 1) & ~3);
	ret = -EFAULT;
	if (copy_to_user(current_dir, cdir, len))
		goto out_free_kenv;

	if ((ret = sys_chdir(current_dir))) {
		goto out_free_kenv;
	}

	ret = -EINVAL;
	if (str2unistr(nls, &current_dir_uni, current_dir))
		goto out_free_kenv;

	if (str2unistr(nls, &image_name_uni, image_name))
		goto out_free_cwd;

	if (native_app) {
		str2unistr(nls, &image_name_uni, native_app);
		kfree(native_app);
	}

	if (desktop && !desktop_info) {
		str2unistr(nls, &desktop_info_uni, desktop);
		kfree(desktop);
	}

	current_dir_handle = NULL;
	console_handle = NULL;
	console_flags = 0;

	len = sizeof(RTL_USER_PROCESS_PARAMETERS)	/* size of process parameter block */
		+ MAX_PATH * sizeof(wchar_t)	/* size of current directory buffer */
		+ ALIGN_TO_LONG(dll_path_uni.MaximumLength)
		+ ALIGN_TO_LONG(image_name_uni.MaximumLength)
		+ ALIGN_TO_LONG(cmd_line_uni.MaximumLength)
		+ ALIGN_TO_LONG(window_title_uni.MaximumLength)
		+ ALIGN_TO_LONG(desktop_info_uni.MaximumLength)
		+ ALIGN_TO_LONG(shell_info_uni.MaximumLength)
		+ ALIGN_TO_LONG(rt_info_uni.MaximumLength)
		+ ALIGN_TO_LONG(env_size);
	/* FIXME: just for pass the socket fd, may be unused in the future */
	len += sizeof(int);

	/* Calculate the required block size */
	brk_size = ALIGN(len, PAGE_SIZE);

	addr = 0x20000000UL - brk_size;
	down_write(&current->mm->mmap_sem);
	brk_res = do_brk(addr, brk_size);
	up_write(&current->mm->mmap_sem);
	if (brk_res != addr)
		goto out_free_image;

	ppb = (PRTL_USER_PROCESS_PARAMETERS)addr;
	memset(ppb, 0, sizeof(*ppb));
	ppb->AllocationSize = brk_size;
	ppb->Size = len;
	ppb->Flags = PPF_NORMALIZED;
	ppb->CurrentDirectoryHandle = current_dir_handle;
	ppb->hConsole = console_handle;
	ppb->ProcessGroup = console_flags;

	ptr = (void *)(ppb + 1);

	/* FIXME: just for pass the socket fd, may be unused in the future */
	memset(ppb + 1, 0, sizeof(int));
	ptr += 4;

	/* copy current directory */
	ret = copy_param_str(ptr, &ppb->CurrentDirectoryName, &current_dir_uni);
	if (ret < 0)
		goto out_free_image;

	/* copy dll path */
	ptr += ret;
	ret = copy_param_str(ptr, &ppb->DllPath, &dll_path_uni);
	if (ret < 0)
		goto out_free_image;

	/* copy image path name */
	ptr += ret;
	ret = copy_param_str(ptr, &ppb->ImagePathName, &image_name_uni);
	if (ret < 0)
		goto out_free_image;

	/* copy command line */
	ptr += ret;
	ret = copy_param_str(ptr, &ppb->CommandLine, &cmd_line_uni);
	if (ret < 0)
		goto out_free_image;

	/* copy title */
	ptr += ret;
	ret = copy_param_str(ptr, &ppb->WindowTitle, &window_title_uni);
	if (ret < 0)
		goto out_free_image;

	/* copy desktop */
	ptr += ret;
	ret = copy_param_str(ptr, &ppb->DesktopInfo, &desktop_info_uni);
	if (ret < 0)
		goto out_free_image;

	/* copy shell info */
	ptr += ret;
	ret = copy_param_str(ptr, &ppb->ShellInfo, &shell_info_uni);
	if (ret < 0)
		goto out_free_image;

	/* copy runtime info */
	ptr += ret;
	ret = copy_param_str(ptr, &ppb->RuntimeInfo, &rt_info_uni);
	if (ret < 0)
		goto out_free_image;

	/* copy Environment */
	ptr += ret;
	ret = copy_to_user(ppb->Environment = ptr, kenv, env_size);
	if (ret)
		goto out_free_image;

	*ppb_res = ppb;
	ret = STATUS_SUCCESS;

out_free_image:
	FREE_UNI(image_name_uni);
out_free_cwd:
	FREE_UNI(current_dir_uni);
out_free_kenv:
	kfree(kenv);
out_free_cmd:
	FREE_UNI(cmd_line_uni);
out_free_envp:
	kfree(envp);
	kfree(argv);
	kfree(cdir);

	return ret;
}

#ifdef SERVER_PM
void server_get_new_process_info(void)
{
	/* in dlls/kernel32/process.c 
	 * RunBuiltinApp() 
	 * and this request will be not used */	
}

struct startup_info *server_get_startup_info(void)
{
	struct w32process *process = get_current_eprocess()->win32process;
	struct startup_info *info = process->startup_info;
	return info;
}
#endif

#endif

